package edu.rice.comp211.laundry.io;

import edu.rice.comp211.laundry.command.*;
import edu.rice.comp211.laundry.garment.Garment;
import edu.rice.comp211.laundry.garment.Pants;
import edu.rice.comp211.laundry.garment.Shirt;
import edu.rice.comp211.laundry.garment.Socks;
import edu.rice.comp211.laundry.iterator.ReadIteratorI;
import edu.rice.comp211.laundry.list.BiListI;
import edu.rice.comp211.laundry.list.EnumI;
import edu.rice.comp211.laundry.student.StudentEnvironment;

import java.io.*;

/**
 * Terminal implementation of the IOProcess interface.
 */
public class TerminalIO implements IOProcess {
   StreamTokenizer in;
   StudentEnvironment sEnv;
   boolean debug;
   InputStream inputStream;
   OutputStream outputStream;

   /* Constructors */
   public TerminalIO(String fName) {
      this(fName, (OutputStream) null);
   }

   public TerminalIO(String inputName, String outputName) {
      this(inputName, getFileOutputStream(outputName));
   }

   public TerminalIO(String inputName, OutputStream os) {
      this(getFileInputStream(inputName), os);
   }

   public TerminalIO(InputStream is, OutputStream os) {
      inputStream = (is == null) ? System.in : is;
      outputStream = (os == null) ? System.out : os;
   }

   private static InputStream getFileInputStream(String inputName) {
      if (inputName != null) {
         try {
            return new FileInputStream(new File(inputName));
         }
         catch (IOException e) {
            throw new RuntimeException(e);
         }
      }
      return null;
   }

   private static OutputStream getFileOutputStream(String outputName) {
      if (outputName != null) {
         try {
            return new PrintStream(new File(outputName));
         }
         catch (IOException e) {
            throw new RuntimeException(e);
         }
      }
      return null;
   }

   public OutputStream open(StudentEnvironment sEnv, boolean debug) {
      in = new StreamTokenizer(new InputStreamReader(inputStream));
      this.sEnv = sEnv;
      this.debug = debug;
      in.resetSyntax();
      in.wordChars('!', '~'); // all printable chars except space, \n, \t, \r
      in.whitespaceChars(0, ' '); // all chars < '|' including space, \n, \t, \r
      return outputStream;
   }

   String parseWord() throws IOException {
      if (in.nextToken() != StreamTokenizer.TT_WORD) {
         throw new CommandException("Non-word encounted as command argument");
      } else {
         return in.sval;
      }
   }

   static Garment parseGarment(String noun, String adjective) throws CommandException {
      //! Use class.forName
      if (noun.equals("shirt")) {
         return new Shirt(adjective);
      } else if (noun.equals("pants")) {
         return new Pants(adjective);
      } else if (noun.equals("socks")) {
         return new Socks(adjective);
      } else {
         throw new CommandException("Illegal garment argument");
      }
   }

   static String spaces = "*   ";
   static String morespaces = "*     ";

//   static private <T> String listToString(EnumI<T> l) {
//      String s = "";  //! should be a StringBuffer
//      ReadIteratorI<T> i = l.newIterator();
//      if (!i.atEnd()) {
//         s = i.currentItem().toString();
//         for (i.next(); !i.atEnd(); i.next()) {
//            s = s + ", " + i.currentItem();
//         }
//      }
//      return s;
//   }

   public Command nextCommand() throws IOException {
      if (debug) {
         System.err.println("* " + sEnv.name() + "'s environment is:");
         System.err.println(spaces + "Outfit: " + sEnv.shirt() +
               ", " + sEnv.pants() + ", " + sEnv.socks());
         System.err.println(spaces + "Closet shelf: ");
         System.err.println(morespaces + sEnv.cleanShirts());
         System.err.println(morespaces + sEnv.cleanPants());
         System.err.println(morespaces + sEnv.cleanSocks());
         System.err.println(spaces + "Dirty pile: ");
         System.err.println(morespaces + sEnv.dirtyPile());
         System.err.println(spaces + "Laundry Room: ");
         EnumI<BiListI<Garment>> ll = sEnv.laundryRoom();

         for (ReadIteratorI<BiListI<Garment>> i = ll.newIterator(); !i.atEnd(); i.next()) {
            System.err.println(morespaces + i.currentItem());
         }
         System.err.println("");
      }
      Command cmd = getCommand();
      if (debug) {
         if (cmd == null) {
            System.err.println("End of File reached");
         } else {
            System.err.println("Command is: " + cmd);
         }
      }
      return cmd;
   }

   private Command getCommand() throws IOException {
      // return null on end-of-file
      int word = in.nextToken();

      if (word == StreamTokenizer.TT_EOF) {
         return null;
      } else if (word == StreamTokenizer.TT_WORD) {

         // Recognize commands with no arguments
         String cmdName = in.sval;
         if (cmdName.equals("launder")) {
            return new Launder();
         } else if (cmdName.equals("fold")) {
            return new Fold();
         } else if (cmdName.equals("outfit")) {
            return new Outfit();
         }

         // Recognize commands with arguments
         else if (cmdName.equals("change")) {
            return new Change(parseGarment(parseWord(), ""));
         } else if (cmdName.equals("lose") || cmdName.equals("receive")) {
            String adjective = parseWord();
            String gmtName = parseWord();
            if (cmdName.equals("lose")) {
               return new Lose(parseGarment(gmtName, adjective));
            } else {
               return new Receive(parseGarment(gmtName, adjective));
            }
         } else {
            throw new CommandException("Illegal command");
         }
      } else {
         throw new CommandException("Illegal command");
      }
   }

   public static void main(String[] args) throws IOException {
      final IOProcess io = new TerminalIO(null);
      io.open(null, false);
      Command c = io.nextCommand();
      while (c != null) {
         System.out.println("Parsed command: " + c);
         c = io.nextCommand();
      }
   }
}