package edu.rice.comp211.laundry.student;

import edu.rice.comp211.laundry.command.Command;
import edu.rice.comp211.laundry.command.CommandVisitor;
import edu.rice.comp211.laundry.command.DoCommandVisitor;
import edu.rice.comp211.laundry.garment.Garment;
import edu.rice.comp211.laundry.garment.Pants;
import edu.rice.comp211.laundry.garment.Shirt;
import edu.rice.comp211.laundry.garment.Socks;
import edu.rice.comp211.laundry.io.IOProcess;
import edu.rice.comp211.laundry.list.*;


import java.io.IOException;
import java.io.PrintStream;

/**
 * Represents a student and their environment
 * 
 *
 */
public class Student implements StudentEnvironment {

	/**
	 * The name of the student
	 */
	private final String name;

	/**
	 * The shirt the student is currently wearing
	 */
	private Shirt shirt;           
	
	/**
	 * The pants the student is currently wearing
	 */
	private Pants pants;
	
	/**
	 * The socks the student is currently wearing
	 */
	private Socks socks;

	/**
	 * The ordered pile of clean shirts the student has available to wear
	 */
	private final BiListI<Garment> cleanShirts; 
	
	/**
	 * The ordered pile of clean socks the student has available to wear
	 */
	private final BiListI<Garment> cleanSocks;
	
	/**
	 * The ordered pile of clean pants the student has available to wear
	 */
	private final BiListI<Garment> cleanPants;
	
	/**
	 * The ordered pile of dirty clothes the student has already worn
	 */
	private final BiListI<Garment> dirtyPile;
	
	/**
	 * A list of ordered piles (lists) of clean clothes that have been 
	 * laundered but not yet returned to the clean clothes piles
	 */
	private final BiListI<BiListI<Garment>> laundryRoom;

	/**
	 * Constructor for the class
	 * @param name the student's name
	 * @param shirt the initial shirt the student is wearing
	 * @param pants the initial pants the student is wearing
	 * @param socks the initial socks the student is wearing
	 */
	public Student(String name, Shirt shirt, Pants pants, Socks socks) {
		this(name, shirt, pants, socks, new BiList<Garment>(), new BiList<Garment>(), new BiList<Garment>(), new BiList<Garment>(), new BiList<BiListI<Garment>>());
	}

	/**
	 * Constructor for the class
	 * @param name the student's name
	 * @param shirt the initial shirt the student is wearing
	 * @param pants the initial pants the student is wearing
	 * @param socks the initial socks the student is wearing
	 * @param cleanShirts the list to use as the clean shirts pile
	 * @param cleanPants the list to use as the clean pants pile
	 * @param cleanSocks the list to use as the clean socks pile
	 * @param dirtyPile the list to use as the dirty clothes pile
	 * @param laundryRoom the list of lists to use as the laundry room
	 */
	public Student(String name, Shirt shirt, Pants pants, Socks socks, BiListI<Garment> cleanShirts,
			BiListI<Garment> cleanPants, BiListI<Garment> cleanSocks, BiListI<Garment> dirtyPile,
			BiListI<BiListI<Garment>> laundryRoom) {
		this.name = name;
		this.shirt = shirt;
		this.pants = pants;
		this.socks = socks;
		this.cleanShirts = cleanShirts;
		this.cleanPants = cleanPants;
		this.cleanSocks = cleanSocks;
		this.dirtyPile = dirtyPile;
		this.laundryRoom = laundryRoom;
	}


	/**
	 * Accessor for the student's name
	 * @return the student's name
	 */
	public String name() {
		return name;
	}

	/**
	 * The current shirt the student is wearing
	 * @return the student's current shirt
	 */
	public Shirt shirt() {
		return shirt;
	}

	/**
	 * The current pants the student is wearing
	 * @return the student's current pants
	 */
	public Pants pants() {
		return pants;
	}

	/**
	 * The current socks the student is wearing
	 * @return the student's current socks
	 */
	public Socks socks() {
		return socks;
	}

	/**
	 * Settor for the current shirt the student is wearing
	 * @param pShirt the new shirt to wear
	 */
	 public void shirt(final Shirt pShirt) {
		shirt = pShirt;
	 }

		/**
		 * Settor for the current pants the student is wearing
		 * @param pPants the new pants to wear
		 */
	 public void pants(final Pants pPants) {
		 pants = pPants;
	 }

		/**
		 * Settor for the current socks the student is wearing
		 * @param pSocks the new socks to wear
		 */
	 public void socks(final Socks pSocks) {
		 socks = pSocks;
	 }

	 /**
	  * Accessor for the pile of clean shirts
	  * @return the clean shirts pile
	  */
	 public BiListI<Garment> cleanShirts() {
		 return cleanShirts;
	 }

	 /**
	  * Accessor for the pile of clean pants
	  * @return the clean pants pile
	  */
	 public BiListI<Garment> cleanPants() {
		 return cleanPants;
	 }

	 /**
	  * Accessor for the pile of clean socks
	  * @return the clean socks pile
	  */
	 public BiListI<Garment> cleanSocks() {
		 return cleanSocks;
	 }

	 /**
	  * Accessor for the pile of dirty clothes
	  * @return the dirty clothes pile
	  */
	 public BiListI<Garment> dirtyPile() {
		 return dirtyPile;
	 }

	 /**
	  * Accessor for the laundry room
	  * @return the laundry room list of lists. 
	  */
	 public BiListI<BiListI<Garment>> laundryRoom() {
		 return laundryRoom;
	 }

	 /**
	  * Tests for equality with another object
	  * @return true if the other object is a student with equal name, clothes and all clothes piles. false otherwise.
	  */
	 public boolean equals(Object o) {
		 if (o == null || o.getClass() != getClass()) {
			 return false;
		 }
		 Student s = (Student) o;
		 return shirt.equals(s.shirt) && pants.equals(s.pants) && socks.equals(s.socks) && cleanShirts.equals(s.cleanShirts)
		 && cleanPants.equals(s.cleanPants) && cleanSocks.equals(s.cleanSocks) && dirtyPile.equals(s.dirtyPile) &&
		 laundryRoom.equals(s.laundryRoom);
	 }

	 /**
	  * a string representation of the student
	  * @return a string with the student's information
	  */
	 public String toString() {
		 return "Student(" + name + ". " + shirt + ", " + pants + ", " + socks + ", " + cleanShirts + ", " + cleanPants
		 + ", " + cleanSocks + ", " + dirtyPile + ", " + laundryRoom + ")";
	 }

	 /**
	  * Run a simulation (a series of commands) on the student
	  * @param io the IOProcess that supplies commands as well as status display capabilities
	  * @param debug a flag used for debugging
	  * @throws IOException if an I/O error occurs.
	  */
	 public void simulate(final IOProcess io, final boolean debug) throws IOException {
		 // holds printstream for simulation

		 PrintStream pOut = new PrintStream(io.open(this, debug));

		 final CommandVisitor<String> cv = getCommandVisitor();

		 Command cmd = io.nextCommand();
		 while (cmd != null) {
			 String status = cmd.accept(cv);
			 pOut.println(status);
			 cmd = io.nextCommand();
		 }
	 }

	 /**
	  * Executes a single command cmd, updating the state of student and generating specified
	  * PrintStream output to System.out.
	  * @param cmd the Command object to execute
	  * @param io used to print the resultant status string.
	  */
	 public void execute(Command cmd, IOProcess io) {
		 PrintStream pOut = new PrintStream(io.open(this, true));
		 CommandVisitor<String> cv = getCommandVisitor();//io, true);
		 pOut.println(cmd.accept(cv));
	 }


	 /**
	  * Factory method for a CommandVisitor&lt;String&gt; that can be used to process Commands for this student
	  * 
	  * @return a DoCommandVisitor instance
	  */
	 public CommandVisitor<String> getCommandVisitor(){ 
		 return  new DoCommandVisitor(this);
	 }
}
